/*
 * Line6 Linux USB driver - 0.7.4
 *
 * Copyright (C) 2004-2008 Markus Grabner (grabner@icg.tugraz.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include "driver.h"

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include "audio.h"
#include "pcm.h"
#include "pod.h"


/* hardware definition (capture) */
static struct snd_pcm_hardware snd_pod_capture_hw = {
	.info = (SNDRV_PCM_INFO_MMAP |
					 SNDRV_PCM_INFO_INTERLEAVED |
					 SNDRV_PCM_INFO_BLOCK_TRANSFER |
					 SNDRV_PCM_INFO_MMAP_VALID |
					 SNDRV_PCM_INFO_SYNC_START),
	.formats =          SNDRV_PCM_FMTBIT_S24_3LE,
	.rates =            SNDRV_PCM_RATE_KNOT,
	.rate_min =         39062,
	.rate_max =         39063,
	.channels_min =     2,
	.channels_max =     2,
	.buffer_bytes_max = 60000,
	.period_bytes_min = POD_ISO_PACKET_SIZE_MAX * POD_BYTES_PER_FRAME,  /* at least one URB must fit into one period */
	.period_bytes_max = 32768,
	.periods_min =      1,
	.periods_max =      1024,
};


/*
	Find a free URB and submit it.
*/
static int submit_audio_in_urb(struct snd_pcm_substream *substream)
{
	int index;
	unsigned long flags;
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);
	int i, urb_size;
	struct urb *urb_in;

	spin_lock_irqsave(&chip->lock_audio_in, flags);
	index = find_first_zero_bit(&chip->active_urb_in, POD_ISO_BUFFERS);

	if(index < 0 || index >= POD_ISO_BUFFERS) {
		spin_unlock_irqrestore(&chip->lock_audio_in, flags);
		dev_err(s2m(substream), "no free URB found\n");
		return -EINVAL;
	}

	urb_in = chip->urb_audio_in[index];
	urb_size = 0;

	for(i = 0; i < POD_ISO_PACKETS; ++i) {
		struct usb_iso_packet_descriptor *fin = &urb_in->iso_frame_desc[i];
		fin->offset = urb_size;
		fin->length = chip->max_packet_size;
		urb_size += chip->max_packet_size;
	}

	urb_in->transfer_buffer = chip->buffer_in + index * POD_ISO_PACKETS * chip->max_packet_size;
	urb_in->transfer_buffer_length = urb_size;
	urb_in->context = substream;

	if(usb_submit_urb(urb_in, GFP_ATOMIC) == 0)
		set_bit(index, &chip->active_urb_in);
	else
		dev_err(s2m(substream), "URB in #%d submission failed\n", index);

	spin_unlock_irqrestore(&chip->lock_audio_in, flags);
	return 0;
}

/*
	Submit all currently available capture URBs.
*/
static int submit_audio_in_all_urbs(struct snd_pcm_substream *substream)
{
	int ret, i;

	for(i = 0; i < POD_ISO_BUFFERS; ++i)
		if((ret = submit_audio_in_urb(substream)) < 0)
			return ret;

	return 0;
}

/*
	Unlink all currently active capture URBs.
*/
static void unlink_audio_in_urbs(struct snd_pod_pcm *chip)
{
	unsigned int i;

	for(i = POD_ISO_BUFFERS; i--;) {
		if(test_bit(i, &chip->active_urb_in)) {
			if(!test_and_set_bit(i, &chip->unlink_urb_in)) {
				struct urb *u = chip->urb_audio_in[i];
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 14)
				u->transfer_flags |= URB_ASYNC_UNLINK;
#endif
				usb_unlink_urb(u);
			}
		}
	}
}

/*
	Wait until unlinking of all currently active capture URBs has been finished.
*/
static void wait_clear_audio_in_urbs(struct snd_pod_pcm *chip)
{
	int timeout = HZ;
	unsigned int i;
	int alive;

	do {
		alive = 0;
		for (i = POD_ISO_BUFFERS; i--;) {
			if (test_bit(i, &chip->active_urb_in))
				alive++;
		}
		if (! alive)
			break;
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(1);
	} while (--timeout > 0);
	if (alive)
		snd_printk(KERN_ERR "timeout: still %d active urbs..\n", alive);

	chip->active_urb_in = 0;
	chip->unlink_urb_in = 0;
}

/*
	Unlink all currently active capture URBs, and wait for finishing.
*/
void unlink_wait_clear_audio_in_urbs(struct snd_pod_pcm *chip)
{
	unlink_audio_in_urbs(chip);
	wait_clear_audio_in_urbs(chip);
}

/*
	Callback for completed capture URB.
*/
static void audio_in_callback(struct urb *urb PT_REGS)
{
	int i, index, length = 0, shutdown = 0;
	int frames;
	unsigned long flags;

	struct snd_pcm_substream *substream = (struct snd_pcm_substream *)urb->context;
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	/* find index of URB */
	for(index = 0; index < POD_ISO_BUFFERS; ++index)
		if(urb == chip->urb_audio_in[index])
			break;

	spin_lock_irqsave(&chip->lock_audio_in, flags);

	for(i = 0; i < POD_ISO_PACKETS; ++i) {
		char *fbuf;
		int fsize;
		struct usb_iso_packet_descriptor *fin = &urb->iso_frame_desc[i];

		if(fin->status == -18) {
			shutdown = 1;
			break;
		}

		fbuf = urb->transfer_buffer + fin->offset;
		fsize = fin->actual_length;
		length += fsize;

		if(fsize > 0) {
			frames = fsize / POD_BYTES_PER_FRAME;

			if(chip->pos_in_done + frames > runtime->buffer_size) {
				/*
					The transferred area goes over buffer boundary,
					copy two separate chunks.
				*/
				int len;
				len = runtime->buffer_size - chip->pos_in_done;

				if(len > 0) {
					memcpy(runtime->dma_area + chip->pos_in_done * POD_BYTES_PER_FRAME, fbuf, len * POD_BYTES_PER_FRAME);
					memcpy(runtime->dma_area, fbuf + len * POD_BYTES_PER_FRAME, (frames - len) * POD_BYTES_PER_FRAME);
				}
				else
					dev_err(s2m(substream), "driver bug: len = %d\n", len);  /* this is somewhat paranoid */
			}
			else {
				/* copy single chunk */
				memcpy(runtime->dma_area + chip->pos_in_done * POD_BYTES_PER_FRAME, fbuf, fsize * POD_BYTES_PER_FRAME);
			}

			if((chip->pos_in_done += frames) >= runtime->buffer_size)
				chip->pos_in_done -= runtime->buffer_size;
		}
	}

	clear_bit(index, &chip->active_urb_in);

	if(test_bit(index, &chip->unlink_urb_in))
		shutdown = 1;

	spin_unlock_irqrestore(&chip->lock_audio_in, flags);

	if(!shutdown) {
		submit_audio_in_urb(substream);

		if((chip->bytes_in += length) >= chip->period_in) {
			chip->bytes_in -= chip->period_in;
			snd_pcm_period_elapsed(substream);
		}
	}
}

/* open capture callback */
static int snd_pod_capture_open(struct snd_pcm_substream *substream)
{
	int err;
	struct snd_pcm_runtime *runtime = substream->runtime;

	if((err = snd_pcm_hw_constraint_ratdens(runtime, 0, SNDRV_PCM_HW_PARAM_RATE, &pod_rates)) < 0)
		return err;

	runtime->hw = snd_pod_capture_hw;
	return 0;
}

/* close capture callback */
static int snd_pod_capture_close(struct snd_pcm_substream *substream)
{
	return 0;
}

/* hw_params capture callback */
static int snd_pod_capture_hw_params(struct snd_pcm_substream *substream, struct snd_pcm_hw_params *hw_params)
{
	int ret;
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);

	/* -- Florian Demski [FD] */
	/* don't ask me why, but this fixes the bug on my machine */
	if ( chip == NULL ) {
		if ( substream->pcm == NULL )
			return -ENOMEM;
		if ( substream->pcm->private_data == NULL )
			return -ENOMEM;
		substream->private_data = substream->pcm->private_data;
		chip = snd_pcm_substream_chip( substream );
	}
	/* -- [FD] end */

	if((ret = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params))) < 0)
		return ret;

	chip->period_in = params_period_bytes(hw_params);
	chip->buffer_in = kmalloc(POD_ISO_BUFFERS * POD_ISO_PACKETS * POD_ISO_PACKET_SIZE_MAX, GFP_KERNEL);

	if(!chip->buffer_in) {
		dev_err(s2m(substream), "cannot malloc buffer_in\n");
		return -ENOMEM;
	}

	return 0;
}

/* hw_free capture callback */
static int snd_pod_capture_hw_free(struct snd_pcm_substream *substream)
{
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);
	unlink_wait_clear_audio_in_urbs(chip);

	if(chip->buffer_in) {
		kfree(chip->buffer_in);
		chip->buffer_in = 0;
	}

	return snd_pcm_lib_free_pages(substream);
}

/* trigger callback */
int snd_pod_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);
	int err;
	chip->count_in = 0;

	switch(cmd) {
	case SNDRV_PCM_TRIGGER_START:
		if(!test_and_set_bit(BIT_RUNNING_CAPTURE, &chip->flags)) {
			err = submit_audio_in_all_urbs(substream);

			if(err < 0) {
				clear_bit(BIT_RUNNING_CAPTURE, &chip->flags);
				return err;
			}
		}

		break;

	case SNDRV_PCM_TRIGGER_STOP:
		if(test_and_clear_bit(BIT_RUNNING_CAPTURE, &chip->flags))
			unlink_audio_in_urbs(chip);

		break;

	default:
		return -EINVAL;
	}

	return 0;
}

/* capture pointer callback */
static snd_pcm_uframes_t
snd_pod_capture_pointer(struct snd_pcm_substream *substream)
{
	struct snd_pod_pcm *chip = snd_pcm_substream_chip(substream);
	return chip->pos_in_done;
}

/* capture operators */
struct snd_pcm_ops snd_pod_capture_ops = {
	.open =        snd_pod_capture_open,
	.close =       snd_pod_capture_close,
	.ioctl =       snd_pcm_lib_ioctl,
	.hw_params =   snd_pod_capture_hw_params,
	.hw_free =     snd_pod_capture_hw_free,
	.prepare =     snd_pod_prepare,
	.trigger =     snd_pod_trigger,
	.pointer =     snd_pod_capture_pointer,
};

int create_audio_in_urbs(struct snd_pod_pcm *chip)
{
	int i;
	struct usb_line6_pod *pod = chip->pod;

	/* create audio URBs and fill in constant values: */
	for(i = 0; i < POD_ISO_BUFFERS; ++i) {
		struct urb *urb;

		/* URB for audio in: */
		urb = chip->urb_audio_in[i] = usb_alloc_urb(POD_ISO_PACKETS, GFP_KERNEL);

		if(urb == NULL) {
			dev_err(pod->line6.ifcdev, "Out of memory\n");
			return -ENOMEM;
		}

		urb->dev = pod->line6.usbdev;
		urb->pipe = usb_rcvisocpipe(pod->line6.usbdev, chip->ep_audio_read & USB_ENDPOINT_NUMBER_MASK);
		urb->transfer_flags = URB_ISO_ASAP;
		urb->start_frame = -1;
		urb->number_of_packets = POD_ISO_PACKETS;
		urb->interval = POD_ISO_INTERVAL;
		urb->error_count = 0;
		urb->complete = audio_in_callback;
	}

	return 0;
}
